/*
 * Copyright (C) 2009 Mukund Sivaraman.
 *
 * Public-domain inflate routine from zpipe.c by Mark Adler.
 *
 * This software is distributed using the modified BSD license, a copy
 * of which you can find in the LICENSE file distributed with this
 * software.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>
#include <zlib.h>
#include "pendian.h"

#define CHUNK 16384

#ifndef PATH_MAX
#define PATH_MAX 256
#endif

struct filerec {
  short unknown1;
  unsigned long long compressed_size;
  unsigned long long uncompressed_size;
  unsigned long long filename_len;
  char unknown2;
  char md5[16];
} __attribute__((packed));

/* Decompress from file source to file dest until stream ends or EOF.
   inf() returns Z_OK on success, Z_MEM_ERROR if memory could not be
   allocated for processing, Z_DATA_ERROR if the deflate data is
   invalid or incomplete, Z_VERSION_ERROR if the version of zlib.h and
   the version of the library linked do not match, or Z_ERRNO if there
   is an error reading or writing the files. */
static int
inf(FILE *source, FILE *dest)
{
  int ret;
  unsigned have;
  z_stream strm;
  unsigned char in[CHUNK];
  unsigned char out[CHUNK];

  /* allocate inflate state */
  strm.zalloc = Z_NULL;
  strm.zfree = Z_NULL;
  strm.opaque = Z_NULL;
  strm.avail_in = 0;
  strm.next_in = Z_NULL;
  ret = inflateInit(&strm);
  if (ret != Z_OK)
    return ret;

  /* decompress until deflate stream ends or end of file */
  do {
    strm.avail_in = fread(in, 1, CHUNK, source);
    if (ferror(source)) {
      (void)inflateEnd(&strm);
      return Z_ERRNO;
    }
    if (strm.avail_in == 0)
      break;
    strm.next_in = in;

    /* run inflate() on input until output buffer not full */
    do {
      strm.avail_out = CHUNK;
      strm.next_out = out;
      ret = inflate(&strm, Z_NO_FLUSH);
      assert(ret != Z_STREAM_ERROR);  /* state not clobbered */
      switch (ret) {
      case Z_NEED_DICT:
        ret = Z_DATA_ERROR;     /* and fall through */
      case Z_DATA_ERROR:
      case Z_MEM_ERROR:
        (void)inflateEnd(&strm);
        return ret;
      }
      have = CHUNK - strm.avail_out;
      if (fwrite(out, 1, have, dest) != have || ferror(dest)) {
        (void)inflateEnd(&strm);
        return Z_ERRNO;
      }
    } while (strm.avail_out == 0);

    /* done when inflate() says it's done */
  } while (ret != Z_STREAM_END);

  /* clean up and return */
  (void)inflateEnd(&strm);
  return ret == Z_STREAM_END ? Z_OK : Z_DATA_ERROR;
}

static int
extract_file (FILE *in)
{
  int ret = 1;
  char *filename = NULL;

  do {
    struct filerec rec;
    size_t n;
    FILE *out;

    n = fread (&rec, 1, sizeof rec, in);
    if (n != sizeof rec) {
      fprintf (stderr, "Error reading from input file\n");
      break;
    }

    rec.unknown1 = be16toh (rec.unknown1);
    rec.compressed_size = be64toh (rec.compressed_size);
    rec.uncompressed_size = be64toh (rec.uncompressed_size);
    rec.filename_len = be64toh (rec.filename_len);

    if ((rec.unknown1 != 0x0001) ||
        (rec.filename_len > PATH_MAX)) {
      fprintf (stderr, "Error reading from input file\n");
      break;
    }

    filename = (char *) malloc (rec.filename_len + 1);

    n = fread (filename, 1, rec.filename_len, in);
    if (n != rec.filename_len)  {
      fprintf (stderr, "Error reading from input file\n");
      break;
    }

    filename [rec.filename_len] = 0;

    printf ("Extracting: %s\n", filename);

    out = fopen (filename, "wb");
    if (!out) {
      fprintf (stderr, "Error opening output file '%s'\n",
               filename);
      break;
    }

    if (inf (in, out) == Z_OK)
      ret = 0;

    fclose (out);

  } while (0);

  if (filename)
    free (filename);

  return ret;
}

int
main (int argc, char *argv[])
{
  int ret = EXIT_FAILURE;
  FILE *in = NULL;

  if (argc != 2) {
    printf ("Usage: %s <zix-file>\n", argv[0]);
    return EXIT_SUCCESS;
  }

  do {
    char header[8];
    size_t n;

    in = fopen (argv[1], "rb");
    if (!in) {
      fprintf (stderr, "Error opening input file '%s'\n",
               argv[1]);
      break;
    }

    n = fread (header, 1, sizeof header, in);
    if (n != sizeof header)  {
      fprintf (stderr, "No ZIX file in file '%s'\n",
               argv[1]);
      break;
    }

    if (strncmp (header, "WINZIX", 6)) {
      fprintf (stderr, "No ZIX file in '%s'\n",
               argv[1]);
      break;
    }

    if ((header[6] != 0x00) || (header[7] != 0x03)) {
      fprintf (stderr, "Unsupported ZIX file in '%s'\n",
               argv[1]);
      break;
    }

    while (!feof (in)) {
      if (extract_file (in))
        break;
    }

    ret = EXIT_SUCCESS;

  } while (0);

  if (in)
    fclose (in);

  return ret;
}
